import sys
import os
import json
from glob import glob
import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
#%matplotlib inline

environment_list = [
"BabyAI-GoToRedBallNoDists-v0",
#"MiniGrid-BlockedUnlockPickup-v0", #NACE only
"MiniGrid-DistShift2-v0",
"MiniGrid-DoorKey-8x8-v0",
"MiniGrid-Empty-16x16-v0",
#"MiniGrid-Empty-5x5-v0", #TODO NACE
"MiniGrid-Empty-6x6-v0",
"MiniGrid-Empty-8x8-v0",
"MiniGrid-Empty-Random-5x5-v0",
"MiniGrid-Empty-Random-6x6-v0",
#MiniGrid-FourRooms-v0 #NOT GOOD WITH ANY
"MiniGrid-LavaCrossingS11N5-v0",
"MiniGrid-LavaGapS7-v0",
#MiniGrid-MultiRoom-N6-v0 #NOT GOOD WITH ANY
"MiniGrid-SimpleCrossingS11N5-v0",
"MiniGrid-Unlock-v0",
"MiniGrid-UnlockPickup-v0"]

base_dir='./'
save_dir='./SavedCSV/'
algos=['TRPO', 'PPO', 'A2C', 'DQN', 'NACE']
window = 1000
C = 0.5

def Plot2_perEnv_all_in_one_all_seeds(base_dir, save_dir, envname, algos, window, C):
    # Create a 2x2 subplot layout
    fig, axs = plt.subplots(1, 2, figsize=(10, 5))
    # Iterate over available algorithms
    for i, algo in enumerate(algos):
        df = pd.read_csv(f"{save_dir}{envname}_{algo}.csv")

        # Plot Left: Mean Episode Reward
        rolling_ep_reward_MEAN = df['Mean Episode Reward'].rolling(window, min_periods=1).mean()
        rolling_ep_reward_STD  = df['Standard Deviation of Episode Reward'].rolling(window, min_periods=1).mean()
        axs[0].set_xscale('log')
        axs[0].plot(df['Timesteps'], rolling_ep_reward_MEAN, label=algo)
        axs[0].fill_between(df['Timesteps'],
                            rolling_ep_reward_MEAN - C*rolling_ep_reward_STD,
                            rolling_ep_reward_MEAN + C*rolling_ep_reward_STD,
                            alpha=0.2,)
        # Plot Right: Mean Episode Length
        rolling_ep_length_MEAN = df['Mean Episode Length'].rolling(window, min_periods=1).mean()
        rolling_ep_length_STD  = df['Standard Deviation of Episode Length'].rolling(window, min_periods=1).mean()
        axs[1].set_xscale('log')
        axs[1].plot(df['Timesteps'], rolling_ep_length_MEAN, label=algo)
        axs[1].fill_between(df['Timesteps'],
                            rolling_ep_length_MEAN - C*rolling_ep_length_STD,
                            rolling_ep_length_MEAN + C*rolling_ep_length_STD,
                           alpha=0.2,)

    # Titles and Legends
    axs[0].set_title('Mean Episode Reward')
    axs[0].set_xlabel('Timesteps')
    axs[0].set_ylabel('Mean Reward')
    axs[1].set_title('Mean Episode Length')
    axs[1].set_xlabel('Timesteps')
    axs[1].set_ylabel('Mean Length')
    # Adjust layout to prevent overlap
    fig.suptitle(envname + " (Evaluation)")
    plt.tight_layout()
    plt.legend()
    # Show the plot
    #plt.show(block=False)
    dire= base_dir + "Figure/"
    if not os.path.exists(dire):
        os.makedirs(dire)
    fig.savefig(dire+f"{envname}.png", dpi=144, format=None, metadata=None, bbox_inches=None,
                pad_inches=0.0, facecolor='auto', edgecolor='auto',backend=None)

seeds=range(5)
for envname in environment_list:
    Plot2_perEnv_all_in_one_all_seeds(base_dir, save_dir, envname, algos, window, C)
